@@ -1,5 +1,5 @@
 #
-# Copyright (c), 2018-2020, SISSA (International School for Advanced Studies).
+# Copyright (c), 2018-2021, SISSA (International School for Advanced Studies).
 # All rights reserved.
 # This file is distributed under the terms of the MIT License.
 # See the file 'LICENSE' in the root directory of the present
@@ -19,13 +19,13 @@ from ..datatypes import AnyAtomicType, AbstractDateTime, Duration, DayTimeDurati
     xsd10_atomic_types, xsd11_atomic_types, ATOMIC_VALUES
 from ..xpath_context import XPathSchemaContext
 from ..tdop import Parser
-from ..namespaces import XSD_ANY_SIMPLE_TYPE, XML_NAMESPACE, XSD_NAMESPACE, \
-    XSD_UNTYPED_ATOMIC, get_namespace, get_expanded_name, split_expanded_name
+from ..namespaces import XSD_ANY_SIMPLE_TYPE, XSD_ANY_ATOMIC_TYPE, XML_NAMESPACE, \
+    XSD_NAMESPACE, XSD_UNTYPED_ATOMIC, get_namespace, get_expanded_name, split_expanded_name
 from ..schema_proxy import AbstractSchemaProxy
 from ..xpath_token import XPathToken
-from ..xpath_nodes import AttributeNode, TypedAttribute, TypedElement, \
-    is_etree_element, is_xpath_node, match_element_node, is_schema_node, \
-    is_document_node, match_attribute_node, is_element_node, node_kind
+from ..xpath_nodes import XPathNode, TypedElement, AttributeNode, TypedAttribute, \
+    is_xpath_node, match_element_node, is_schema_node, is_document_node, \
+    match_attribute_node, is_element_node, node_kind
 
 
 OPERATORS_MAP = {
@@ -175,7 +175,7 @@ class XPath1Parser(Parser):
             self.parser.advance('(')
             if nargs is None:
                 del self[:]
-                if self.parser.next_token.symbol == ')':
+                if self.parser.next_token.symbol in (')', '(end)'):
                     raise self.error(code, 'at least an argument is required')
                 while True:
                     self.append(self.parser.expression(5))
@@ -186,7 +186,9 @@ class XPath1Parser(Parser):
                 return self
             elif nargs == 0:
                 if self.parser.next_token.symbol != ')':
-                    raise self.error(code, '%s has no arguments' % str(self))
+                    if self.parser.next_token.symbol != '(end)':
+                        raise self.error(code, '%s has no arguments' % str(self))
+                    raise self.parser.next_token.wrong_syntax()
                 self.parser.advance()
                 return self
             elif isinstance(nargs, (tuple, list)):
@@ -196,7 +198,7 @@ class XPath1Parser(Parser):
 
             k = 0
             while k < min_args:
-                if self.parser.next_token.symbol == ')':
+                if self.parser.next_token.symbol in (')', '(end)'):
                     msg = 'Too few arguments: expected at least %s arguments' % min_args
                     raise self.wrong_nargs(msg if min_args > 1 else msg[:-1])
 
@@ -215,7 +217,7 @@ class XPath1Parser(Parser):
                 elif k == 0 and self.parser.next_token.symbol != ')':
                     self[k:] = self.parser.expression(5),
                 else:
-                    break
+                    break  # pragma: no cover
                 k += 1
 
             if self.parser.next_token.symbol == ',':
@@ -260,8 +262,12 @@ class XPath1Parser(Parser):
         if get_namespace(type_qname) == XSD_NAMESPACE:
             if type_qname == XSD_UNTYPED_ATOMIC:
                 return isinstance(obj, UntypedAtomic)
+            elif type_qname == XSD_ANY_ATOMIC_TYPE:
+                return isinstance(obj, AnyAtomicType)
             elif type_qname == XSD_ANY_SIMPLE_TYPE:
-                return isinstance(obj, (AnyAtomicType, list))
+                return isinstance(obj, AnyAtomicType) or \
+                    isinstance(obj, list) and \
+                    all(isinstance(x, AnyAtomicType) for x in obj)
 
             try:
                 if self.xsd_version == '1.1':
@@ -271,7 +277,11 @@ class XPath1Parser(Parser):
                 pass
 
         if self.schema is not None:
-            return self.schema.is_instance(obj, type_qname)
+            try:
+                return self.schema.is_instance(obj, type_qname)
+            except KeyError:
+                pass
+
         raise ElementPathKeyError("unknown type %r" % type_qname)
 
     def is_sequence_type(self, value):
@@ -460,22 +470,28 @@ def select(self, context=None):
         # Try to match the type using the item's path
         for item in context.iter_children_or_self():
             if match_attribute_node(item, name) or match_element_node(item, tag):
-                path = context.get_path(item)
-
-                xsd_node = self.parser.schema.find(path, self.parser.namespaces)
-                if xsd_node is not None:
-                    self.xsd_types = {tag: xsd_node.type}
+                if isinstance(item, (TypedAttribute, TypedElement)):
+                    yield item
                 else:
-                    self.xsd_types = self.parser.schema
+                    path = context.get_path(item)
 
-                context.item = self.get_typed_node(item)
-                yield context.item
+                    xsd_node = self.parser.schema.find(path, self.parser.namespaces)
+                    if xsd_node is not None:
+                        self.xsd_types = {tag: xsd_node.type}
+                    else:
+                        self.xsd_types = self.parser.schema
+
+                    context.item = self.get_typed_node(item)
+                    yield context.item
     else:
         # XSD typed selection
         for item in context.iter_children_or_self():
             if match_attribute_node(item, name) or match_element_node(item, tag):
-                context.item = self.get_typed_node(item)
-                yield context.item
+                if isinstance(item, (TypedAttribute, TypedElement)):
+                    yield item
+                else:
+                    context.item = self.get_typed_node(item)
+                    yield context.item
 
 
 ###
@@ -552,22 +568,28 @@ def select(self, context=None):
     elif self.xsd_types is None or isinstance(self.xsd_types, AbstractSchemaProxy):
         for item in context.iter_children_or_self():
             if match_attribute_node(item, name) or match_element_node(item, name):
-                path = context.get_path(item)
-                xsd_node = self.parser.schema.find(path, self.parser.namespaces)
-                if xsd_node is not None:
-                    self.add_xsd_type(xsd_node)
+                if isinstance(item, (TypedAttribute, TypedElement)):
+                    yield item
                 else:
-                    self.xsd_types = self.parser.schema
+                    path = context.get_path(item)
+                    xsd_node = self.parser.schema.find(path, self.parser.namespaces)
+                    if xsd_node is not None:
+                        self.add_xsd_type(xsd_node)
+                    else:
+                        self.xsd_types = self.parser.schema
 
-                context.item = self.get_typed_node(item)
-                yield context.item
+                    context.item = self.get_typed_node(item)
+                    yield context.item
 
     else:
         # XSD typed selection
         for item in context.iter_children_or_self():
             if match_attribute_node(item, name) or match_element_node(item, name):
-                context.item = self.get_typed_node(item)
-                yield context.item
+                if isinstance(item, (TypedAttribute, TypedElement)):
+                    yield item
+                else:
+                    context.item = self.get_typed_node(item)
+                    yield context.item
 
 
 ###
@@ -611,15 +633,18 @@ def select(self, context=None):
     elif self.xsd_types is None:
         for item in context.iter_children_or_self():
             if match_attribute_node(item, name):
-                yield item[-1]
+                yield item.value
             elif match_element_node(item, name):
                 yield item
     else:
         # XSD typed selection
         for item in context.iter_children_or_self():
             if match_attribute_node(item, name) or match_element_node(item, name):
-                context.item = self.get_typed_node(item)
-                yield context.item
+                if isinstance(item, (TypedAttribute, TypedElement)):
+                    yield item
+                else:
+                    context.item = self.get_typed_node(item)
+                    yield context.item
 
 
 ###
@@ -671,18 +696,19 @@ def select(self, context=None):
                 pass  # '*' wildcard doesn't match document nodes
             elif context.axis == 'attribute':
                 if isinstance(item, (AttributeNode, TypedAttribute)):
-                    yield item[-1]
+                    yield item.value
             elif is_element_node(item):
                 yield item
 
     else:
         # XSD typed selection
         for item in context.iter_children_or_self():
-            if context.item is None:
-                pass
-            elif context.is_principal_node_kind():
-                context.item = self.get_typed_node(item)
-                yield context.item
+            if context.item is not None and context.is_principal_node_kind():
+                if isinstance(item, (TypedAttribute, TypedElement)):
+                    yield item
+                else:
+                    context.item = self.get_typed_node(item)
+                    yield context.item
 
 
 @method(nullary('.'))
@@ -978,12 +1004,16 @@ def select(self, context=None):
                 raise self.error('XPTY0019')
 
             for result in self[1].select(context):
-                if not is_etree_element(result) and not isinstance(result, tuple):
+                if not isinstance(result, (tuple, XPathNode)) and not hasattr(result, 'tag'):
                     yield result
                 elif result in items:
                     pass
-                elif isinstance(result, (TypedAttribute, TypedElement)):
-                    if result[0] not in items:
+                elif isinstance(result, TypedElement):
+                    if result.elem not in items:
+                        items.add(result)
+                        yield result
+                elif isinstance(result, TypedAttribute):
+                    if result.attribute not in items:
                         items.add(result)
                         yield result
                 else:
