@@ -1,5 +1,5 @@
 #
-# Copyright (c), 2018-2020, SISSA (International School for Advanced Studies).
+# Copyright (c), 2018-2021, SISSA (International School for Advanced Studies).
 # All rights reserved.
 # This file is distributed under the terms of the MIT License.
 # See the file 'LICENSE' in the root directory of the present
@@ -448,6 +448,7 @@ unregister('lang')
 unregister('id')
 unregister('substring-before')
 unregister('substring-after')
+unregister('starts-with')
 
 ###
 # Symbols
@@ -1224,7 +1225,7 @@ def select(self, context=None):
                 elif not type_name:
                     yield attribute.value
                 elif isinstance(attribute, TypedAttribute):
-                    if attribute.type.name == type_name:
+                    if attribute.xsd_type.name == type_name:
                         yield attribute.value
                 else:
                     xsd_type = self.get_xsd_type(attribute)
